"""This file contains Logic related to the .metadata file."""

import json
import tempfile
import typing
from pathlib import Path
from typing import Optional

import msgspec
import threadful
from packaging.requirements import InvalidRequirement, Requirement
from result import Err, Ok, Result

from ._maybe import Empty, Maybe
from ._python import _pip, _uv
from ._symlinks import check_symlinks

if typing.TYPE_CHECKING:
    from typing_extensions import Self

T = typing.TypeVar("T", bound=typing.Any)
V = typing.TypeVar("V", bound=typing.Any)


class Metadata(msgspec.Struct, array_like=True):
    """Structure of the .metadata file."""

    name: str
    scripts: dict[str, bool]  # {script: is_installed}
    install_spec: str  # e.g. '2fas' or '2fas[gui]>=0.1.0'
    extras: set[str]  # .e.g. {'gui'}
    requested_version: Optional[str]  # e.g. ">=0.1.0"
    installed_version: str
    python: str = ""
    python_raw: str = ""
    injected: Optional[set[str]] = None

    @typing.overload
    def _convert_type(self, value: set[V]) -> list[V]:  # type: ignore
        """Convert set of V to list of V."""

    @typing.overload
    def _convert_type(self, value: T) -> T:
        """Other types are not affected."""

    def _convert_type(self, value: set[V] | T) -> list[V] | T:
        """Make the value JSON-encodable."""
        if isinstance(value, set):
            return list(value)
        return value

    def to_dict(self) -> dict[str, typing.Any]:
        """Convert the struct to a JSON-safe dictionary."""
        return {f: self._convert_type(getattr(self, f)) for f in self.__struct_fields__}

    def check_script_symlinks(self, name: str) -> "Self":
        """Update self.scripts with the current status of these scripts' symlinks."""
        self.scripts = check_symlinks(self.scripts.keys(), venv=name)
        return self


encoder = msgspec.msgpack.Encoder()
decoder = msgspec.msgpack.Decoder(type=Metadata)


def fake_install(spec: str) -> dict:
    """Dry run pip to extract metadata of a local package."""
    _uv("pip", "install", "pip")  # ensure we have pip

    with tempfile.NamedTemporaryFile() as f:
        _pip("install", "--no-deps", "--dry-run", "--ignore-installed", "--report", f.name, spec)

        return json.load(f)


@threadful.thread
def resolve_local(spec: str) -> tuple[Maybe[str], Maybe[str]]:
    """Resolve the package name of a local package by dry run installing it."""
    try:
        full_data = fake_install(spec)
        install_data = full_data["install"][0]

        name = install_data["metadata"]["name"]
        extras = install_data.get("requested_extras")
        file_url = install_data["download_info"]["url"]

        if extras:
            _extras = ",".join(extras)
            return Ok(f"{name}[{_extras}]"), Ok(file_url)
        else:
            return Ok(name), Ok(file_url)
    except Exception:
        return Empty(), Empty()


def collect_metadata(spec: str) -> Result[Metadata, InvalidRequirement]:
    """Parse an install spec into a (incomplete) Metadata object."""
    try:
        parsed_spec = Requirement(spec)

    except InvalidRequirement as e:
        local, path = threadful.animate(resolve_local(spec), text=f"Trying to install local package '{spec}'")

        match local, path:
            case Ok(_local), Ok(_path):
                # both ok:
                parsed_spec = Requirement(_local)
                spec = f"{parsed_spec.name} @ {_path.removeprefix('file://')}"
            case _, _:
                # any err:
                return Err(e)

    return Ok(
        Metadata(
            install_spec=spec,
            name=parsed_spec.name,
            scripts={},  # postponed
            extras=parsed_spec.extras,
            requested_version=str(parsed_spec.specifier),
            installed_version="",  # postponed
            python="",  # postponed
        )
    )


def store_metadata(meta: Metadata, venv: Path):
    """Save the metadata struct to .metadata in a venv."""
    with (venv / ".metadata").open("wb") as f:
        f.write(encoder.encode(meta))


def read_metadata(venv: Path) -> Maybe[Metadata]:
    """Read the .metadata file for a venv."""
    metafile = venv / ".metadata"
    if not metafile.exists():
        return Empty()

    with metafile.open("rb") as f:
        data = decoder.decode(f.read())  # type: Metadata
        return Ok(data)



import sys
import textwrap
from pathlib import Path

import plumbum  # type: ignore
from plumbum.cmd import grep  # type: ignore
from plumbum.machines.local import LocalCommand  # type: ignore
from uv import find_uv_bin

_python = plumbum.local[sys.executable]
_pip = _python["-m", "pip"]
_uv = plumbum.local[find_uv_bin()]


def _python_in_venv(venv: Path) -> LocalCommand:
    python = venv / "bin" / "python"
    return plumbum.local[python]


def _uv_in_venv(venv: Path):
    python = _python_in_venv(venv)
    return python["-m", "uv"]


def _run_python_in_venv(*args: str, venv: Path) -> str:
    python = _python_in_venv(venv)

    return python(*args)


def run_python_code_in_venv(code: str, venv: Path) -> str:
    """
    Run Python code in a virtual environment.

    Args:
        code (str): The Python code to run.
        venv (Path): The path of the virtual environment.

    Returns:
        str: The output of the Python code.
    """
    code = textwrap.dedent(code)
    return _run_python_in_venv("-c", code, venv=venv)


def get_python_version(venv: Path):
    """Get the reported (human-readable) version of Python installed in a venv."""
    return _run_python_in_venv("--version", venv=venv).strip()


def get_python_executable(venv: Path):
    """Get the Python executable for a venv (used to determine the version)."""
    executable = venv / "bin" / "python"  # DON'T use _python_in_venv because we want to resolve the symlink:
    return str(executable.resolve())  # /usr/bin/python3.xx


def get_package_version(package: str, venv: Path) -> str:
    """Get the currently installed version of a specific package."""
    # assumes `with virtualenv(venv)` block executing this function
    uv = _uv_in_venv(venv)

    regex = f"^({package}==|{package} @)"
    line: str = (uv["pip", "freeze"] | grep["-E", regex])().strip()
    return (line.split("@")[-1] if "@" in line else line.split("==")[-1]).strip()



import typing

from result import Err, Ok

T = typing.TypeVar("T")


class Empty(Err[None]):
    """
    Alias for Err[None], used by Maybe.

    Usage:
    Result[str, None] = Maybe[str]
    """

    def __init__(self) -> None:
        """Set up the Err result with None as it's value."""
        super().__init__(None)


Maybe = Ok[T] | Empty



"""Core functionality."""

import shutil
import sys
import typing
from contextlib import contextmanager
from pathlib import Path
from typing import Optional

import plumbum  # type: ignore
import rich
from plumbum import local  # type: ignore
from result import Err, Ok, Result
from threadful import thread
from threadful.bonus import animate

from ._constants import WORK_DIR
from ._maybe import Maybe
from ._python import _uv, get_package_version, get_python_executable, get_python_version
from ._symlinks import find_symlinks, install_symlinks, remove_symlink
from .metadata import Metadata, collect_metadata, read_metadata, store_metadata


@thread
def uv(*args: typing.Any, **kwargs: str):
    """
    Execute the uv command with the provided arguments.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Raises:
        NotImplementedError: If kwargs is not empty.

    Returns:
        str: The result of the uv command.
    """
    if kwargs:
        raise NotImplementedError("todo")

    result: str = _uv(*args)

    return result


@contextmanager
def virtualenv(virtualenv_dir: Path | str) -> typing.Generator[Path, None, None]:
    # https://github.com/tomerfiliba/plumbum/issues/168
    """
    Context manager for executing commands within the context of a Python virtualenv.

    Args:
        virtualenv_dir (Path | str): The directory of the virtual environment.

    Yields:
        Path: The path of the virtual environment.
    """
    old_path = local.env["PATH"]

    if not isinstance(virtualenv_dir, Path):
        virtualenv_dir = Path(virtualenv_dir)

    virtualenv_bin_dir = str((virtualenv_dir / "bin").resolve())
    new_path = "{}:{}".format(virtualenv_bin_dir, old_path)
    old_env = local.env.get("VIRTUAL_ENV")
    new_env = str(virtualenv_dir)
    local.env["PATH"] = new_path
    local.env["VIRTUAL_ENV"] = new_env
    old_python = local.python
    new_python = local["python"]
    local.python = new_python
    try:
        yield virtualenv_dir
    finally:
        local.env["PATH"] = old_path
        local.env["VIRTUAL_ENV"] = old_env
        local.python = old_python


@contextmanager
def as_virtualenv(venv_name: str) -> typing.Generator[Path, None, None]:
    """
    External variant of the 'virtualenv' context manager above.

    This variant takes in a venv name instead of a whole path.
    """
    workdir = ensure_local_folder()
    venv_path = workdir / "venvs" / venv_name
    with virtualenv(venv_path):
        yield venv_path


@contextmanager
def exit_on_pb_error() -> typing.Generator[None, None, None]:
    """Pass the plumbum error to the stderr and quit."""
    try:
        yield
    except plumbum.ProcessExecutionError as e:
        print(e.stderr, file=sys.stderr)
        exit(e.retcode)


def format_bools(data: dict[str, bool], sep=" | ") -> str:
    """Given a dictionary, format every format with the value True in green and False in red."""
    return sep.join([f"[green]{k}[/green]" if v else f"[red]{k}[/red]" for k, v in data.items()])


def install_package(
    package_name: str,
    venv: Optional[Path] = None,
    python: Optional[str] = None,
    force: bool = False,
    extras: Optional[typing.Iterable[str]] = None,
    no_cache: bool = False,
) -> Result[str, Exception]:
    """
    Install a package in a virtual environment.

    Args:
        package_name (str): The name of the package.
        venv (Optional[Path], optional): The path of the virtual environment. Defaults to None.
        python (str): version or executable to use.
        force (bool, optional): If True, overwrites existing package. Defaults to False.
        extras: which optional features ('extras') to install for this package.
        no_cache: don't use `uv` cache.
    """
    if extras is None:
        extras = []

    match collect_metadata(package_name):
        case Err(e):
            return Err(e)
        case Ok(meta):
            # just bind meta
            ...

    if venv is None:
        venv = create_venv(meta.name, python=python, force=force)

    with virtualenv(venv), exit_on_pb_error():
        args: list[str] = []
        text = f"installing {meta.name}"
        if extras:
            args.extend(extras)
            text += f" with {extras}"

        if no_cache or force:
            args += ["--no-cache"]

        try:
            animate(uv("pip", "install", meta.install_spec, *args), text=text)

            # must still be in the venv and try for these:
            meta.installed_version = get_package_version(meta.name, venv)
            meta.python = get_python_version(venv)
            meta.python_raw = get_python_executable(venv)
            meta.injected = set(extras)

        except plumbum.ProcessExecutionError as e:
            remove_dir(venv)
            return Err(e)

    msg = ""
    if install_symlinks(meta.name, venv, force=force, meta=meta):
        msg = f"📦 {meta.name} ({meta.installed_version}) installed!"  # :package:

    store_metadata(meta, venv)
    return Ok(msg)


def reinstall_package(
    package_name: str,
    python: Optional[str] = None,
    force: bool = False,
    with_injected: bool = True,
    no_cache: bool = False,
) -> Result[str, Exception]:
    """
    Reinstalls a package in a virtual environment.

    This function first collects metadata about the package, then checks if the package is already installed in the
    virtual environment. If the package is not installed, it prompts the user to install it first. If the package is
    installed, it uninstalls the package and then reinstalls it. If a new version or extra is requested or no old
    metadata is available, it installs from the CLI argument package name. Otherwise, it installs from the old metadata
    spec. The Python version used for the virtual environment can be specified. If not specified, it uses the Python
    version from the existing metadata if available.

    Args:
        package_name (str): The name of the package to reinstall.
        python (Optional[str], optional): The Python version to use for the virtual environment. Defaults to None.
        force (bool, optional): If True, ignores if the virtual environment does not exist. Defaults to False.
        with_injected (bool): also re-include injected packages?
        no_cache: don't use `uv` cache.

    Raises:
        SystemExit: If the package is not installed in the virtual environment and force is False.
    """
    match collect_metadata(package_name):
        case Err(e):
            # can't work without metadata, just stop
            return Err(e)
        case Ok(new_metadata):
            # bind new_metadata
            ...

    workdir = ensure_local_folder()
    venv = workdir / "venvs" / new_metadata.name

    if not venv.exists() and not force:
        return Err(
            ValueError(
                f"'{new_metadata.name}' was not previously installed. Please run 'uvx install {package_name}' instead."
            )
        )

    existing_metadata = read_metadata(venv)

    # if a new version or extra is requested or no old metadata is available, install from cli arg package name.
    # otherwise, install from old metadata spec
    new_install_spec = bool(new_metadata.requested_version or new_metadata.extras)

    # install_spec = package_name if new_install_spec or not existing_metadata else existing_metadata.install_spec
    metadata: Optional[Metadata] = None
    match (new_install_spec, existing_metadata):
        case (False, Ok(metadata)):
            install_spec = metadata.install_spec
        case _:
            # if new install spec is True or there is no old metadata:
            install_spec = package_name

    # python = python or (existing_metadata.python_raw if existing_metadata else None)
    python = python or existing_metadata.map_or(None, lambda m: m.python_raw)

    uninstall_package(new_metadata.name, force=force)
    extras = metadata.injected if (with_injected and metadata and metadata.injected) else []
    return install_package(install_spec, python=python, force=force, extras=extras, no_cache=no_cache)


def inject_packages(into: str, package_specs: set[str]) -> Result[str, Exception]:
    """Install extra libraries into a package-specific venv."""
    match collect_metadata(into):
        case Err(e):
            return Err(e)
        case Ok(meta):
            # just bind meta
            ...

    workdir = ensure_local_folder()
    venv = workdir / "venvs" / meta.name

    if not venv.exists():
        return Err(ValueError(f"'{meta.name}' was not previously installed. Please run 'uvx install {into}' first."))

    # after getting the right venv, load actual metadata (with fallback to previous emptier metadata):
    meta = read_metadata(venv).unwrap_or(meta)

    with virtualenv(venv), exit_on_pb_error():
        try:
            animate(uv("pip", "install", *package_specs), text=f"injecting {package_specs}")
        except plumbum.ProcessExecutionError as e:
            return Err(e)

    meta.injected = (meta.injected or set()) | package_specs
    store_metadata(meta, venv)

    return Ok(f"💉 Injected {package_specs} into {meta.name}.")  # :needle:


def remove_dir(path: Path):
    """
    Remove a directory.

    Args:
        path (Path): The path of the directory.
    """
    if path.exists():
        shutil.rmtree(path)


def upgrade_package(
    package_name: str, force: bool = False, skip_injected: bool = False, no_cache: bool = False
) -> Result[str, Exception]:
    """Upgrade a package in its venv."""
    # run `uv pip install --upgrade package` with requested install spec (version, extras, injected)
    # if --force is used, the previous version is ignored.
    match collect_metadata(package_name):
        case Err(e):
            return Err(e)
        case Ok(spec_metadata):
            # bind spec_metadata
            ...

    workdir = ensure_local_folder()
    venv = workdir / "venvs" / spec_metadata.name

    if not venv.exists():
        return Err(NotADirectoryError(f"No virtualenv for '{package_name}', stopping. Use 'uvx install' instead."))

    meta = read_metadata(venv).unwrap_or(spec_metadata)

    old_version = meta.installed_version

    with virtualenv(venv), exit_on_pb_error():
        # pip upgrade package[extras]==version *injected
        # if version spec in spec_metadata use that instead
        # if --force, drop version spec
        base_pkg = meta.name
        extras = meta.extras

        injected: set[str] = (not skip_injected and meta.injected) or set()
        # injected = set() if skip_injected else (meta.injected or set())

        version: str = spec_metadata.requested_version or (not force and meta.requested_version) or ""
        # version = spec_metadata.requested_version or ("" if force else meta.requested_version)

        options = []
        if force:
            options.append("--no-cache")

        upgrade_spec = base_pkg + version
        if extras:
            upgrade_spec += "[" + ",".join(extras) + "]"

        try:
            animate(uv("pip", "install", "--upgrade", upgrade_spec, *injected, *options), text=f"upgrading {base_pkg}")
        except plumbum.ProcessExecutionError as e:
            return Err(e)

        meta.requested_version = version
        new_version = meta.installed_version = get_package_version(meta.name, venv)
        store_metadata(meta, venv)

    if old_version == new_version:
        msg = f"🌟 '{package_name}' is already up to date at version {new_version}!"
        if meta.requested_version:
            msg += (f"\n💡 This package was installed with a version constraint ({meta.requested_version}). "
                    f"If you want to ignore this constraint, use `uvx upgrade --force {package_name}`.")

    else:
        msg = f"🚀 Successfully updated '{package_name}' from version {old_version} to version {new_version}!"

    return Ok(msg)


def uninstall_package(package_name: str, force: bool = False) -> Result[str, Exception]:
    """
    Uninstalls a package.

    Args:
        package_name (str): The name of the package.
        force (bool, optional): If True, ignores if the virtual environment does not exist. Defaults to False.
    """
    workdir = ensure_local_folder()
    venv_path = workdir / "venvs" / package_name

    if not venv_path.exists() and not force:
        escaped = package_name.replace("[", "\\[")
        return Err(
            NotADirectoryError(
                f"No virtualenv for '{escaped}', stopping. "
                f"Use '--force' to remove an executable with that name anyway.",
            )
        )

    meta = read_metadata(venv_path)

    symlinks = find_symlinks(package_name, venv_path) or [package_name]

    for symlink in symlinks:
        remove_symlink(symlink)

    remove_dir(venv_path)

    _version = meta.map_or("", lambda m: f" ({m.installed_version})")
    return Ok(_version)


def ensure_local_folder() -> Path:
    """
    Ensure the local folder exists.

    Returns:
        Path: The path of the local folder.
    """
    (WORK_DIR / "venvs").mkdir(exist_ok=True, parents=True)
    return WORK_DIR


def create_venv(name: str, python: Optional[str] = None, force: bool = False, with_pip: bool = True) -> Path:
    """
    Create a virtual environment.

    Args:
        name (str): The name of the virtual environment.
        python (str): which version of Python to use (e.g. 3.11, python3.11)
        force (bool): ignore existing venv
        with_pip (bool): also install (regular) pip into the venv? (required for `uvx runpip`)

    Returns:
        Path: The path of the virtual environment.
    """
    workdir = ensure_local_folder()

    venv_path = workdir / "venvs" / name

    if venv_path.exists() and not force:
        rich.print(
            f"'{name}' is already installed. "
            f"Use 'uvx upgrade' to update existing tools or pass '--force' to this command to ignore this message.",
            file=sys.stderr,
        )
        exit(1)

    args = ["venv", venv_path]

    if python:
        args.extend(["--python", python])

    uv(*args).join()

    if with_pip:
        with virtualenv(venv_path):
            animate(uv("pip", "install", "pip", "uv"))

    return venv_path


def list_packages() -> typing.Generator[tuple[str, Maybe[Metadata]], None, None]:
    """Iterate through all uvx venvs and load the metadata files one-by-one."""
    workdir = ensure_local_folder()

    for subdir in workdir.glob("venvs/*"):
        metadata = read_metadata(subdir)

        yield subdir.name, metadata


def run_command(command: str, *args: str, printfn: typing.Callable[..., None] = print) -> int:
    """Run a command via plumbum without raising an error on exception."""
    # retcode = None makes the command not raise an exception on error:
    exit_code, stdout, stderr = plumbum.local[command][args].run(retcode=None)

    if stdout:
        printfn(stdout.strip())
    if stderr:
        printfn(stderr.strip(), file=sys.stderr)
    return exit_code



"""Symlink-related logic."""

import sys
import typing
from pathlib import Path
from typing import Optional

from ._constants import BIN_DIR, WORK_DIR
from ._python import run_python_code_in_venv

if typing.TYPE_CHECKING:
    from .metadata import Metadata


def install_symlink(symlink: str, venv: Path, force: bool = False, binaries: tuple[str, ...] = ()) -> bool:
    """
    Install a symlink in the virtual environment.

    Args:
        symlink (str): The name of the symlink.
        venv (Path): The path of the virtual environment.
        force (bool, optional): If True, overwrites existing symlink. Defaults to False.
        binaries (tuple[str, ...], optional): The binaries to install. Defaults to ().

    Returns:
        bool: True if the symlink was installed, False otherwise.
    """
    if binaries and symlink not in binaries:
        return False

    target_path = BIN_DIR / symlink

    if target_path.exists():
        if force:
            target_path.unlink()
        else:
            print(
                f"Script {symlink} already exists in {BIN_DIR}. Use --force to ignore this warning.",
                file=sys.stderr,
            )
            return False

    symlink_path = venv / "bin" / symlink
    if not symlink_path.exists():
        print(
            f"Could not symlink {symlink_path} because the script didn't exist.",
            file=sys.stderr,
        )
        return False

    target_path.symlink_to(symlink_path)
    return True


def find_symlinks(library: str, venv: Path) -> list[str]:
    """
    Find the symlinks for a library in a virtual environment.

    Args:
        library (str): The name of the library.
        venv (Path): The path of the virtual environment.

    Returns:
        list: The list of symlinks.
    """
    code = f"""
    import importlib.metadata

    for script in importlib.metadata.distribution('{library}').entry_points:
        if script.group != "console_scripts":
            continue

        print(script.name)
    """

    try:
        raw = run_python_code_in_venv(code, venv)
        return [_ for _ in raw.split("\n") if _]
    except Exception:
        return []


def install_symlinks(
    library: str,
    venv: Path,
    force: bool = False,
    binaries: tuple[str, ...] = (),
    meta: "Optional[Metadata]" = None,
) -> bool:
    """
    Install symlinks for a library in a virtual environment.

    Args:
        library (str): The name of the library.
        venv (Path): The path of the virtual environment.
        force (bool, optional): If True, overwrites existing symlinks. Defaults to False.
        binaries (tuple[str, ...], optional): The binaries to install. Defaults to ().
        meta: Optional metadata object to store results in

    Returns:
        bool: True if any symlink was installed, False otherwise.
    """
    symlinks = find_symlinks(library, venv)

    results = {}
    for symlink in symlinks:
        results[symlink] = install_symlink(symlink, venv, force=force, binaries=binaries)

    if meta:
        meta.scripts = results

    return any(results.values())


def check_symlink(symlink: str, venv: str) -> bool:
    """
    Check if a symlink is valid.

    This function checks if a symlink exists and if the target path is within the symlink's resolved parents.
    The symlink is considered valid if both conditions are met.

    Args:
        symlink (str): The name of the symlink to check.
        venv (str): The name of the virtual environment.

    Returns:
        bool: True if the symlink is valid, False otherwise.
    """
    symlink_path = BIN_DIR / symlink
    target_path = WORK_DIR / "venvs" / venv

    return symlink_path.is_symlink() and target_path in symlink_path.resolve().parents


def remove_symlink(symlink: str):
    """
    Remove a symlink.

    Args:
        symlink (str): The name of the symlink.
    """
    target_path = BIN_DIR / symlink
    if target_path.exists() and target_path.is_symlink():
        target_path.unlink(missing_ok=True)


def check_symlinks(symlinks: typing.Iterable[str], venv: str) -> dict[str, bool]:
    """
    Check all symlinks in the input iterable (e.g. list or dict.keys).

    Returns the key and whether the symlink is valid.
    """
    return {k: check_symlink(k, venv) for k in symlinks}



"""This file builds the Typer cli."""

import os
import subprocess  # nosec
import sys
from datetime import datetime
from pathlib import Path
from typing import Annotated

import rich
import typer
from result import Err, Ok, Result
from typer import Context

from uvx._constants import BIN_DIR

from .__about__ import __version__
from ._maybe import Maybe
from ._python import _python_in_venv, _uv
from .core import (
    as_virtualenv,
    format_bools,
    inject_packages,
    install_package,
    list_packages,
    reinstall_package,
    run_command,
    uninstall_package,
    upgrade_package,
)
from .metadata import Metadata

app = typer.Typer()


def output(result: Result[str, Exception]) -> None:
    """Output positive (ok) result to stdout and error result to stderr."""
    match result:
        case Ok(msg):
            rich.print(msg)
        case Err(err):
            rich.print(err, file=sys.stderr)


OPTION_PYTHON_HELP_TEXT = "Python version or executable to use, e.g. `3.12`, `python3.12`, `/usr/bin/python3.12`"


@app.command()
def install(
    package_name: str,
    force: Annotated[
        bool,
        typer.Option(
            "-f", "--force", help="Overwrite currently installed executables with the same name (in ~/.local/bin)"
        ),
    ] = False,
    python: Annotated[str, typer.Option(help=OPTION_PYTHON_HELP_TEXT)] = "",
    no_cache: Annotated[bool, typer.Option("--no-cache", help="Run without `uv` cache")] = False,
):
    """Install a package (by pip name)."""
    output(install_package(package_name, python=python, force=force, no_cache=no_cache))


@app.command(name="upgrade")
@app.command(name="update")
def upgrade(
    package_name: str,
    force: Annotated[bool, typer.Option("-f", "--force", help="Ignore previous version constraint")] = False,
    skip_injected: Annotated[
        bool, typer.Option("--skip-injected", help="Don't also upgrade injected packages")
    ] = False,
    no_cache: Annotated[bool, typer.Option("--no-cache", help="Run without `uv` cache")] = False,
):
    """Upgrade a package."""
    output(upgrade_package(package_name, force=force, skip_injected=skip_injected, no_cache=no_cache))


@app.command(name="remove")
@app.command(name="uninstall")
def uninstall(
    package_name: str,
    force: Annotated[
        bool,
        typer.Option(
            "-f",
            "--force",
            help="Remove executable with the same name (in ~/.local/bin) even if related venv was not found",
        ),
    ] = False,
):
    """Uninstall a package (by pip name)."""
    output(uninstall_package(package_name, force=force).map(lambda version: f"🗑️ {package_name}{version} removed!"))


@app.command()
def reinstall(
    package: str,
    python: Annotated[str, typer.Option(help=OPTION_PYTHON_HELP_TEXT)] = "",
    force: Annotated[bool, typer.Option("-f", "--force", help="See `install --force`")] = False,
    without_injected: Annotated[
        bool, typer.Option("--without-injected", help="Don't include previously injected libraries in reinstall")
    ] = False,
    no_cache: Annotated[bool, typer.Option("--no-cache", help="Run without `uv` cache")] = False,
):
    """Uninstall a package (by pip name) and re-install from the original spec (unless a new spec is supplied)."""
    output(
        reinstall_package(
            package,
            python=python,
            force=force,
            with_injected=not without_injected,
            no_cache=no_cache,
        ).map(lambda _: _.replace(" installed", " reinstalled"))
    )


@app.command()
def inject(into: str, package_specs: list[str]):
    """Install additional packages to a virtual environment managed by uvx."""
    output(
        inject_packages(
            into,
            set(package_specs),
        )
    )


# todo:
# self-upgrade (uv and uvx)
# upgrade-all

@app.command()
def upgrade_all(
    force: Annotated[bool, typer.Option("-f", "--force", help="Ignore previous version constraint")] = False,
    skip_injected: Annotated[
        bool, typer.Option("--skip-injected", help="Don't also upgrade injected packages")
    ] = False,
    no_cache: Annotated[bool, typer.Option("--no-cache", help="Run without `uv` cache")] = False,
):
    """Upgrade all uvx-installed packages."""

    for (venv_name, _) in list_packages():
        upgrade(
            venv_name,
            force=force,
            skip_injected=skip_injected,
            no_cache=no_cache
        )


# list
def _list_short(name: str, metadata: Maybe[Metadata]):
    rich.print("-", name, metadata.map_or("[red]?[/red]", lambda md: md.installed_version))


TAB = " " * 3


def _list_normal(name: str, metadata: Maybe[Metadata], verbose: bool = False):
    match metadata:
        case Err(_):
            print("-", name)
            rich.print(TAB, "[red]Missing metadata [/red]")
            return
        case Ok(md):
            # just binds 'md'
            pass

    extras = list(md.extras)
    name_with_extras = name if not extras else f"{name}{extras}"
    rich.print("-", name_with_extras)

    md.check_script_symlinks(name)

    if verbose:
        rich.print(TAB, str(md))
    else:
        rich.print(
            TAB,
            f"Installed Version: {md.installed_version} on {md.python}.",
        )
        if md.injected:
            p = ", ".join([f"'{_}'" for _ in md.injected])
            rich.print(TAB, f"Injected Packages: {p}")
        rich.print(TAB, "Scripts:", format_bools(md.scripts))


def _list_venvs_json():
    from json import dumps

    print(
        dumps(
            {
                name: metadata.map_or({}, lambda md: md.check_script_symlinks(name).to_dict())
                for name, metadata in list_packages()
            },
            indent=2,
        )
    )


@app.command(name="list")
def list_venvs(short: bool = False, verbose: bool = False, json: bool = False):
    """List packages and apps installed with uvx."""
    if json:
        return _list_venvs_json()

    for name, metadata in list_packages():
        if short:
            _list_short(name, metadata)
        else:
            _list_normal(name, metadata, verbose=verbose)


@app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def runuv(venv: str, ctx: Context):
    """Run 'uv' in the right venv."""
    with as_virtualenv(venv) as venv_path:
        python = _python_in_venv(venv_path)
        run_command(str(python), "-m", "uv", *ctx.args)  # run_command does not work with _uv_in_venv !


@app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def runpip(venv: str, ctx: Context):
    """Run 'pip' in the right venv."""
    with as_virtualenv(venv):
        run_command("pip", *ctx.args)


@app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def runpython(venv: str, ctx: Context):
    """Run 'python' in the right venv."""
    with as_virtualenv(venv) as venv_path:
        python = venv_path / "bin" / "python"
        subprocess.run([python, *ctx.args])  # nosec


def add_to_bashrc(text: str, with_comment: bool = True):
    """Add text to ~/.bashrc, usually with a comment (uvx + timestamp)."""
    with (Path.home() / ".bashrc").resolve().open("a") as f:
        now = str(datetime.now()).split(".")[0]
        final_text = "\n"
        final_text += f"# Added by `uvx` at {now}\n" if with_comment else ""
        final_text += text + "\n"
        f.write(final_text)


@app.command()
def ensurepath(force: bool = False):
    """Update ~/.bashrc with a PATH that includes the local bin directory that uvx uses."""
    env_path = os.getenv("PATH", "")
    bin_in_path = str(BIN_DIR) in env_path.split(":")

    if bin_in_path and not force:
        rich.print(
            f"[yellow]{BIN_DIR} is already added to your path. Use '--force' to add it to your .bashrc file anyway.[/yellow]"
        )
        exit(1)

    add_to_bashrc(f'export PATH="$PATH:{BIN_DIR}"')


@app.command()
def completions():  # noqa
    """
    Use --install-completion to install the autocomplete script, \
        or --show-completion to see what would be installed.
    """
    rich.print("Use 'uvx --install-completion' to install the autocomplete script to your '.bashrc' file.")


def version_callback():
    """Show the current versions when running with --version."""
    rich.print("uvx", __version__)
    run_command(str(_uv), "--version", printfn=rich.print)
    rich.print("Python", sys.version.split(" ")[0])


@app.callback(invoke_without_command=True, no_args_is_help=True)
def main(
    ctx: typer.Context,
    # stops the program:
    version: bool = False,
) -> None:  # noqa
    """
    This callback will run before every command, setting the right global flags.

    Args:
        ctx: context to determine if a subcommand is passed, etc

        version: display current version?

    """
    if version:
        version_callback()
    elif not ctx.invoked_subcommand:
        rich.print("[yellow]Missing subcommand. Try `uvx --help` for more info.[/yellow]")
    # else: just continue


if __name__ == "__main__":  # pragma: no cover
    app()
